---------------------------------------------------------------------------------------------------------
--  ROCKSPAWN SCRIPT
---------------------------------------------------------------------------------------------------------
-- Adds rock spawning capability to cultivators/plows
--
-- Author:		akuenzi (akuenzi@gmail.com)
-- Date:		6/28/2016
--
-- Sincere thanks to Xentro for his patient mentorship in creating this script.
--

source(g_currentModDirectory .. "Rock.lua");

RockSpawn = {};
RockSpawn.ModDir = g_currentModDirectory

-- Rock i3d files
RockSpawn.RockFiles = {
	{"Rocks/Giants_FieldRock01.i3d", fillLevel = 370},
	{"Rocks/Giants_FieldRock02.i3d", fillLevel = 280},
	{"Rocks/Giants_FieldRock03.i3d", fillLevel = 210},
	{"Rocks/Giants_rock1.i3d", fillLevel = 60},
	{"Rocks/TyroSmith_Rock1.i3d", fillLevel = 40},
	{"Rocks/TyroSmith_Rock2.i3d", fillLevel = 30},
	{"Rocks/TyroSmith_Rock3.i3d", fillLevel = 90},
	{"Rocks/TyroSmith_Rock4.i3d", fillLevel = 50},
	{"Rocks/TyroSmith_Rock5.i3d", fillLevel = 50},
	{"Rocks/TyroSmith_Rock6.i3d", fillLevel = 140},
	{"Rocks/Benw11_Rock1.i3d", fillLevel = 90},
	{"Rocks/Benw11_Rock2.i3d", fillLevel = 40},	
	{"Rocks/Benw11_Rock3.i3d", fillLevel = 20},	
	{"Rocks/Benw11_Rock4.i3d", fillLevel = 10},	
	{"Rocks/Giants_FieldRock01g.i3d", fillLevel = 0.4},										-- This is the 'special' rock (gold).  It should always be LAST in this list.		
};

RockSpawn.MAX_NUM_FILES = table.getn(RockSpawn.RockFiles) - 1;		

-- Input Binding for Rock Spawning Levels
RockSpawn.RockLevels = {
	{rlevel = 0, rMinDistance = 0, rMaxDistance = 0},										-- The input binding of '0' is intended to represent the 'off' position.
	{rlevel = 1, rMinDistance = 284.4, rMaxDistance = 853.1},								-- rMinDistance = minimum distance after last rock spawn, that another rock may spawn.
	{rlevel = 2, rMinDistance = 142.2, rMaxDistance = 426.5},								-- rMaxDistance = maximum distance after last rock spawn, that another rock may spawn.
	{rlevel = 3, rMinDistance = 71.1, rMaxDistance = 213.3},
	{rlevel = 4, rMinDistance = 35.5, rMaxDistance = 106.6},
	{rlevel = 5, rMinDistance = 17.8, rMaxDistance = 53.3},	
};

RockSpawn.MAX_LEVELS = table.getn(RockSpawn.RockLevels) - 1;

-- Rock spawn distances
RockSpawn.DEFAULT_WIDTH = 12;																-- Default width of plow/cultivator.  Larger width = less spawning overall.

-- General offset values
RockSpawn.X_OFFSET = 0.50;
RockSpawn.Y_OFFSET = 0.15;																	-- Move spawn point up 
RockSpawn.Z_OFFSET = 0.50; 																	-- Move spawn point backwards
RockSpawn.SEARCH_OFFSET = 0.25;																-- Move raycast 
RockSpawn.MAX_OFFSET = -3;																	-- Maximum Z offset allowed from raycast

function RockSpawn.prerequisitesPresent(specializations)
	return SpecializationUtil.hasSpecialization(Cultivator, specializations) or SpecializationUtil.hasSpecialization(Plough, specializations)
end;

function RockSpawn:load(xmlFile)
	self.setRockSpawnStatus = SpecializationUtil.callSpecializationsFunction("setRockSpawnStatus");	
	self.createRock = RockSpawn.createRock;
	self.ploughOffset = RockSpawn.ploughOffset;
	self.implementWidth = 12																								-- Initial implement width, soon to be changed to actual by the script.
	self.findZ = false
	
	if SpecializationUtil.hasSpecialization(Plough, self.specializations) then
		if hasXMLProperty(xmlFile, "vehicle.ploughTurnSound") then															-- I am a plow!
			self.startIndex = self.workAreas[1].start
			self.scriptCanActivate = self.aiBackMarker ~= nil and self.startIndex ~= nil;
		else
			self.scriptCanActivate = self.aiBackMarker ~= nil and self.aiLeftMarker ~= nil and self.aiRightMarker ~= nil	-- I am a ripper!
		end;
	else
		self.scriptCanActivate = self.aiBackMarker ~= nil and self.aiLeftMarker ~= nil and self.aiRightMarker ~= nil		-- I am a cultivator!
	end;

	if not self.scriptCanActivate then
		print("RockSpawn will not work with this file - [[[",self.configFileName,"]]]  It is missing aiMarkers or startIndex required in order to function.")
	end;
	
	self.RockSpawnStatus = RockSpawn.RockLevels[2]["rlevel"];																-- Default rock spawning level at start:  Level 1
	
	if self.isServer then
		self.cultivatorCollisionRaycast = RockSpawn.cultivatorCollisionRaycast;
		self.searchOffsetDistance = true;
		self.rockRaycastOffset = 0
		self.spawnOffsetZ = nil;
	end;
	
	RockSpawn.resetDistance(self);
end;

function RockSpawn:delete()
end;

function RockSpawn:readStream(streamId, connection)
	self:setRockSpawnStatus(streamReadInt8(streamId), true);
end;

function RockSpawn:writeStream(streamId, connection)
	streamWriteInt8(streamId,self.RockSpawnStatus);
end;

function RockSpawn:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)	
	local status = getXMLInt(xmlFile, key.."#RockSpawnStatus");
	if status ~= nil then
		self:setRockSpawnStatus(status);
	end;
	
	return BaseMission.VEHICLE_LOAD_OK;
end;

function RockSpawn:getSaveAttributesAndNodes(nodeIdent)
	local attributes = 'RockSpawnStatus="' .. self.RockSpawnStatus .. '"';
	return attributes, nil;
end;

function RockSpawn:mouseEvent(posX, posY, isDown, isUp, button)
end;

function RockSpawn:keyEvent(unicode, sym, modifier, isDown)
end;

function RockSpawn:update(dt)
	if self.scriptCanActivate then
		if self:getIsActive() then
			if self:getIsActiveForInput() then
				if InputBinding.hasEvent(InputBinding.RockStatus) then
					local statusChange = self.RockSpawnStatus
					if statusChange == RockSpawn.MAX_LEVELS then
						statusChange = 0
					else
						statusChange = self.RockSpawnStatus + 1
					end;
					
					self:setRockSpawnStatus(statusChange);
					RockSpawn.resetDistance(self);
				end;
			end;
			
			if self.firstTimeRun then
				if self.isServer and self.RockSpawnStatus ~= 0 then
					if self.searchOffsetDistance then
						if self.isLowered and self.doGroundManipulation then
							self.FindZ = true
						elseif self.isLowered and self.ploughHasGroundContact and self.getIsInWorkPosition then
							self.FindZ = true
						else
							self.FindZ = false
						end;
						if self.FindZ then
							local x, y, z = localToWorld(self.aiBackMarker, 0, 2, self.rockRaycastOffset);
							local dx, dy, dz = localDirectionToWorld(self.aiBackMarker, 0, -1, 0);
														
							raycastAll(x, y, z, dx, dy, dz, "cultivatorCollisionRaycast", 5, self);
							
							if self.foundZ ~= nil then
								self.spawnOffsetZ = self.foundZ - RockSpawn.Z_OFFSET;
								self.foundZ = nil;
								self.findZ = false;
								self.searchOffsetDistance = false;
							else
								self.rockRaycastOffset = self.rockRaycastOffset - RockSpawn.SEARCH_OFFSET;
								if 	self.rockRaycastOffset < RockSpawn.MAX_OFFSET then
									self.rockRaycastOffset = 0
								end;
							end;
						end;	
					end;
					
					if self.spawnOffsetZ ~= nil then
						if self.doGroundManipulation or self.ploughHasGroundContact then
							if not self.showFieldNotOwnedWarning then
								if self.traveledDistance >= self.currentSpawnTargetDistance then
									RockSpawn.resetDistance(self);
									self:createRock();
								else
									self.traveledDistance = self.traveledDistance + self.lastMovedDistance;
								end;
							end;	
						end;
					end;
				end;
			end;
		end;
	end;	
end;

function RockSpawn:updateTick(dt)
end;

function RockSpawn:draw()
	if self.scriptCanActivate then
		if self:getIsActiveForInput(true) then
			if self.RockSpawnStatus == 0 then
				g_currentMission:addHelpButtonText(string.format(g_i18n:getText("LEVEL_0"), g_i18n:getText("RockStatus")), InputBinding.RockStatus);

			else
				local helpText
				if self.RockSpawnStatus == 1 then
					helpText = g_i18n:getText("LEVEL_2")
				elseif self.RockSpawnStatus == 3 then
					helpText = g_i18n:getText("LEVEL_3")
				elseif self.RockSpawnStatus == 5 then
					helpText = g_i18n:getText("LEVEL_4")
				else
					helpText = ""
				end;
							
				g_currentMission:addHelpButtonText(string.format(g_i18n:getText("WE_ARE_ACTIVE"), g_i18n:getText("RockStatus"), g_i18n:getText("LEVEL_1"), RockSpawn.RockLevels[self.RockSpawnStatus+1]["rlevel"], helpText), InputBinding.RockStatus);

			end;
		end;	
	end;	
end;

function RockSpawn:createRock()
	assert(self.isServer ~= nil);
	local randomX;
	local rockId = math.ceil(math.random(1, RockSpawn.MAX_NUM_FILES * 100 + 1) / 100);
	if self.startIndex ~= nil then														-- Follow the code below if a plow.
		local Sx, _, Sz = worldToLocal(self.startIndex, 0, 0, 0);
		local Bx, _, Bz = worldToLocal(self.aiBackMarker, 0, 0, 0);
		self.implementWidth = math.abs(Sx - Bx)
		local plowTangent = math.abs(Sz-Bz)/self.implementWidth
		randomX = math.random(self.implementWidth*10000)/10000
		self.additionalZOffset = randomX * plowTangent - RockSpawn.Z_OFFSET
		if Sx > Bx then
			randomX = randomX * -1 - RockSpawn.X_OFFSET
		else
			randomX = randomX + RockSpawn.X_OFFSET
		end;
		
	else																				-- Follow the code below if a cultivator/ripper.
		local a, _, _ = getTranslation(self.aiLeftMarker);
		local d, _, _ = getTranslation(self.aiRightMarker);
		self.implementWidth = math.abs(d - a)
		randomX = math.random(((self.implementWidth/2)*-10000),(self.implementWidth/2)*10000)/10000
		self.additionalZOffset = 0
		
	end;

	local xSpawn, _, zSpawn = localToWorld(self.aiBackMarker, randomX, 0, self.spawnOffsetZ + self.additionalZOffset);
	local ySpawn = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, xSpawn, 0, zSpawn) + RockSpawn.Y_OFFSET;
	
	local rock = Rock:new(self.isServer, self.isClient);
	rock:load(Utils.getFilename(RockSpawn.RockFiles[rockId][1], RockSpawn.ModDir), xSpawn, ySpawn, zSpawn, math.random(0,360), math.random(0,360), math.random(0,360), RockSpawn.RockFiles[rockId]["fillLevel"]);
	rock:register();
end;

function RockSpawn:setRockSpawnStatus(RockSpawnStatus, noEventSend)
	if self.RockSpawnStatus ~= RockSpawnStatus then
		if noEventSend == nil or noEventSend == false then
			if g_server ~= nil then
				g_server:broadcastEvent(setRockSpawnStatusEvent:new(self,RockSpawnStatus),nil,nil,self);
			else
				g_client:getServerConnection():sendEvent(setRockSpawnStatusEvent:new(self,RockSpawnStatus));
			end;
		end;
		
		self.RockSpawnStatus = RockSpawnStatus;
	end;
end;

function RockSpawn:cultivatorCollisionRaycast(transformId, x, y, z, distance)
	local vehicleCollisionObject = g_currentMission.nodeToVehicle[transformId];
	
	if vehicleCollisionObject ~= nil then
		return false;
	elseif transformId == g_currentMission.terrainRootNode then
		local _, _, offsetZ = worldToLocal(self.aiBackMarker, x, y, z);
		self.foundZ = offsetZ;
		return false;
	end; 
	
	return true;
end;

function RockSpawn.resetDistance(self)
	self.traveledDistance = 0;
	self.currentSpawnTargetDistance = math.random(RockSpawn.RockLevels[self.RockSpawnStatus+1]["rMinDistance"]*(RockSpawn.DEFAULT_WIDTH/self.implementWidth), RockSpawn.RockLevels[self.RockSpawnStatus+1]["rMaxDistance"]*(RockSpawn.DEFAULT_WIDTH/self.implementWidth));
end;


---------------------------------------------------------------------------------------------------------
--  SET ROCK SPAWN STATUS EVENT FOR MULTIPLAYER	(ie, WHEN PLOW/CULTIVATOR SPAWNING INPUT BINDING CHANGES)
---------------------------------------------------------------------------------------------------------
setRockSpawnStatusEvent = {};
setRockSpawnStatusEvent_mt = Class(setRockSpawnStatusEvent, Event);

InitEventClass(setRockSpawnStatusEvent, "setRockSpawnStatusEvent");

function  setRockSpawnStatusEvent:emptyNew()
	local self = Event:new(setRockSpawnStatusEvent_mt);
	self.className = "setRockSpawnStatusEvent";
	return self;
end;

function setRockSpawnStatusEvent:new(object,RockSpawnStatus)
	local self = setRockSpawnStatusEvent:emptyNew()
	self.object = object;
	self.RockSpawnStatus = RockSpawnStatus;
	return self;
end;

function setRockSpawnStatusEvent:readStream(streamId, connection)
	self.object = networkGetObject(streamReadInt32(streamId));
	self.RockSpawnStatus = streamReadInt8(streamId);
	self:run(connection);
end;

function setRockSpawnStatusEvent:writeStream(streamId, connection)
	streamWriteInt32(streamId, networkGetObjectId(self.object));
	streamWriteInt8(streamId,self.RockSpawnStatus);
end;

function setRockSpawnStatusEvent:run(connection)
	self.object:setRockSpawnStatus(self.RockSpawnStatus, true);
	if not connection:getIsServer() then
		g_server:broadcastEvent(setRockSpawnStatusEvent:new(self.object,self.RockSpawnStatus),nil,connection,self.object);
	end;
end;